#ifndef INSTRUCTIONS_H
#define INSTRUCTIONS_H

#ifdef  __KERNEL__
#undef  __KERNEL__
#endif
#define __KERNEL__ 1
#include <linux/types.h>
#include <linux/sched.h>

#include "control_w.h"
#include "fpu_emu.h"
#include "fpu_system.h"

extern struct info math_arg;


typedef void (*__longarg_t)(struct info);
typedef __longarg_t longargfunc;


/* fnsave mem; fwait; nop */
#define FNSAVE(x) { \
  static char fsave_code[] = { 0xdd, 0x35, 0x00, 0x00, 0x00, 0x00, 0x9b, 0x90 }; \
			   ((struct fpu_state **)&fsave_code[2])[0] = &x; \
			     math_arg.___eip = (long)fsave_code; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fnstenv mem; fwait; nop */
#define FNSTENV(x) { \
  static char fstenv_code[] = { 0xd9, 0x35, 0x00, 0x00, 0x00, 0x00, 0x9b, 0x90 }; \
			    ((struct fpu_state **)&fstenv_code[2])[0] = &x; \
			     math_arg.___eip = (long)fstenv_code; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fldenv mem; fwait; nop */
#define FLDENV(x) { \
  static char fldenv_code[] = { 0xd9, 0x25, 0x00, 0x00, 0x00, 0x00, 0x9b, 0x90 }; \
			    ((struct fpu_state **)&fldenv_code[2])[0] = &x; \
			     math_arg.___eip = (long)fldenv_code; \
			   ((longargfunc)math_emulate)(math_arg); }

/* frstor mem; fwait; nop */
#define FRSTOR(x) { \
  static char frstor_code[] = { 0xdd, 0x25, 0x00, 0x00, 0x00, 0x00, 0x9b, 0x90 }; \
			    ((struct fpu_state **)&frstor_code[2])[0] = &x; \
			     math_arg.___eip = (long)frstor_code; \
			   ((longargfunc)math_emulate)(math_arg); }

/* frstor mem; nop */
#define FRSTORN(x) { \
static char frstorn_code[] = { 0xdd, 0x25, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
		     ((struct fpu_state **)&frstorn_code[2])[0] = (void *)&x; \
			     math_arg.___eip = (long)frstorn_code; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fnstsw mem; fwait; nop */
#define FNSTSW(x) { \
  static char fstsw_code[] = { 0xdd, 0x3d, 0x00, 0x00, 0x00, 0x00, 0x9b, 0x90 }; \
			   ((unsigned short **)&fstsw_code[2])[0] = &x; \
			     math_arg.___eip = (long)fstsw_code; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fnstsw mem; nop */
#define FNSTSW_(x) { \
  static char fstsw_code_[] = { 0xdd, 0x3d, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
			      ((unsigned short **)&fstsw_code_[2])[0] = &x; \
			     math_arg.___eip = (long)fstsw_code_; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fldcw mem; nop */
#define FLDCW_(x) { \
  static char fldcw_code_[] = { 0xd9, 0x2d, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
			      ((unsigned short **)&fldcw_code_[2])[0] = &x; \
			     math_arg.___eip = (long)fldcw_code_; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fstcw mem; nop */
#define FSTCW_(x) { \
  static char fstcw_code_[] = { 0xd9, 0x3d, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
			      ((unsigned short **)&fstcw_code_[2])[0] = &x; \
			     math_arg.___eip = (long)fstcw_code_; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fldt mem; nop */
#define FLDT(x) { \
  static char fldt_code_[] = { 0xdb, 0x2d, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
			     ((struct long_double **)&fldt_code_[2])[0] = &x; \
			     math_arg.___eip = (long)fldt_code_; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fldl mem; nop */
#define FLDL(x) { \
  static char fldl_code_[] = { 0xdd, 0x05, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
			     ((struct long_double **)&fldl_code_[2])[0] = &x; \
			     math_arg.___eip = (long)fldl_code_; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fstpt mem; nop */
#define FSTPT(x) { \
  static char fstpt_code_[] = { 0xdb, 0x3d, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
		      ((struct long_double **)&fstpt_code_[2])[0] = &x; \
			     math_arg.___eip = (long)fstpt_code_; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fstpl mem; nop */
#define FSTPL(x) { \
  static char fstpl_code_[] = { 0xdd, 0x1d, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
		      ((struct long_double **)&fstpl_code_[2])[0] = &x; \
			     math_arg.___eip = (long)fstpl_code_; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fstl mem; nop */
#define FSTL(x) { \
  static char fstl_code_[] = { 0xdd, 0x15, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
		     ((struct long_double **)&fstl_code_[2])[0] = &x; \
			     math_arg.___eip = (long)fstl_code_; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fsts mem; nop */
#define FSTS(x) { \
  static char fsts_code_[] = { 0xd9, 0x15, 0x00, 0x00, 0x00, 0x00, 0x90 }; \
			     ((struct long_double **)&fsts_code_[2])[0] = &x; \
			     math_arg.___eip = (long)fsts_code_; \
			   ((longargfunc)math_emulate)(math_arg); }

/* fninit; fwait; nop */
#define FNINIT() { \
static char finit_code[] = { 0xdb, 0xe3, 0x9b, 0x90 }; \
		 math_arg.___eip = (long)finit_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fmulp1; nop */
#define FMULP1() { \
  static char fmulp1_code[] = { 0xde, 0xc9, 0x90 }; \
	      math_arg.___eip = (long)fmulp1_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* ftst; nop */
#define FTST_() { \
  static char ftst_code[] = { 0xd9, 0xe4, 0x90 }; \
		    math_arg.___eip = (long)ftst_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fincstp; nop */
#define FINCSTP() { \
  static char fincstp_code[] = { 0xd9, 0xf7, 0x90 }; \
		       math_arg.___eip = (long)fincstp_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fxam; nop */
#define FXAM_() { \
  static char fxam_code[] = { 0xd9, 0xe5, 0x90 }; \
		    math_arg.___eip = (long)fxam_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fwait; nop */
#define FWAIT() { \
  static char fwait_code[] = { 0x9b, 0x90 }; \
	       math_arg.___eip = (long)fwait_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fnclex; nop */
#define FNCLEX() { \
  static char fnclex_code[] = { 0xdb, 0xe2, 0x90 }; \
		      math_arg.___eip = (long)fnclex_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fprem1; nop */
#define FPREM1() { \
  static char fprem1_code[] = { 0xd9, 0xf5, 0x90 }; \
		      math_arg.___eip = (long)fprem1_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fldz; fwait; nop */
#define FLDZ() { \
  static char fldz_code[] = { 0xd9, 0xee, 0x9b, 0x90 }; \
		  math_arg.___eip = (long)fldz_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fld1; fwait; nop */
#define FLD1() { \
  static char fld1_code[] = { 0xd9, 0xe8, 0x9b, 0x90 }; \
		  math_arg.___eip = (long)fld1_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fld1; fadd %st(0),%st; fld %st(0); fwait; nop */
#define TWOTWOS() { \
  static char twotwo_code[] = { 0xd9, 0xe8, 0xd8, 0xc0, 0xd9, 0xc0, 0x9b, 0x90 }; \
		    math_arg.___eip = (long)twotwo_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fmul %st(1),%st; fwait; nop */
#define FMULST() { \
  static char fmulst_code[] = { 0xd8, 0xc9, 0x9b, 0x90 }; \
		    math_arg.___eip = (long)fmulst_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fmul %st(1),%st; nop */
#define FMULSTN() { \
  static char fmulstn_code[] = { 0xd8, 0xc9, 0x90 }; \
		       math_arg.___eip = (long)fmulstn_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fdiv %st(1),%st; fwait; nop */
#define FDIVST() { \
static char fdivst_code[] = { 0xd8, 0xf1, 0x9b, 0x90 }; \
		  math_arg.___eip = (long)fdivst_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fdiv %st(1),%st; nop */
#define FDIVSTN() { \
  static char fdivstn_code[] = { 0xd8, 0xf1, 0x9b, 0x90 }; \
		     math_arg.___eip = (long)fdivstn_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fstp %st(0); fwait; nop */
#define FSTPST0() { \
  static char fstpst0_code[] = { 0xdd, 0xd8, 0x9b, 0x90 }; \
	     math_arg.___eip = (long)fstpst0_code; \
		   ((longargfunc)math_emulate)(math_arg); }

/* fldz; ...(8 times) fwait; nop */
#define FLDZ8() { \
static char fldz8_code[] = { 0xd9, 0xee, 0xd9, 0xee, 0xd9, 0xee, 0xd9, 0xee, \
			    0xd9, 0xee, 0xd9, 0xee, 0xd9, 0xee, 0xd9, 0xee, \
			    0x9b, 0x90 }; \
		    math_arg.___eip = (long)fldz8_code; \
		   ((longargfunc)math_emulate)(math_arg); }

#define MASK_ALL() { unsigned short mem; \
                     FSTCW_(mem); \
                     mem |= 0x7f; \
                     FLDCW_(mem); }

#define OVFL_OFF() { unsigned short mem; \
                     FSTCW_(mem); \
                     mem |= 0x08; \
                     FLDCW_(mem); }

#define OVFL_ON( ) { unsigned short mem; \
                     FSTCW_(mem); \
                     mem &= 0xfff7; \
                     FLDCW_(mem); }

void enable_fpu_interrupts(int interrupts);
unsigned short precision(unsigned short x);

#endif INSTRUCTIONS_H
