/* Test a few special cases of arithmetic */

#define __KERNEL__ 1

#ifndef FPU_EMU_2
#define FPU_EMU_2 1
#endif

#include <linux/types.h>
#include <linux/sched.h>

#include "control_w.h"
#include "status_w.h"
#include "fpu_emu.h"
#include "fpu_system.h"
#include "setup_reg.h"

#include "testio.h"

#include "instructions.h"

#define getsign_(x) (getsign(x) >> 7)

struct task_struct *current;

struct fpu_state everything;

#define NR_CPUS 1
struct task_struct x;
struct task_struct *current_set[NR_CPUS] = { &x };
struct info math_arg;


#ifndef exponent
#define exponent(x)  (*(short *)&((x)->exp))
#endif

#define MUL 0
#define DIV 1
#define ADD 2
#define SUB 3
#define REM1 4
#define SQRT 5
#define SCALE 6
#define SINCOS 7
#define COS 8
#define TAN 9
#define MULP 10
#define ADDI 11
#define REM 12
#define SIN 13

/* Filler */
#define FX 0

#define REGISTER_(sign,exponent,sigl,sigh,tag) \
      { sigl, sigh, (((exponent)+EXTENDED_Ebias) & 0x7fff) | (sign != 0)*0x8000 }, tag

struct test {
  char operation;
  int control;
  FPU_REG arg1; char arg1tag;
  FPU_REG arg2; char arg2tag;
  FPU_REG result; char restag;
  FPU_REG correct0; char c0tag;
  FPU_REG correct1; char c1tag;
  FPU_REG native0; char nat0tag;
  FPU_REG native1; char nat1tag;
  unsigned short swd;
} tests[] = 
  {
    {
      DIV,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_OVER, 0x00000000, 0xc0000100, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_OVER, 0x00000000, 0xc0000100, TAG_Special),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0
    },
    {
      MUL,
      PR_53_BITS | RC_UP,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000800, 0x80000001, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000800, 0xffffffff, TAG_Valid),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( SIGN_POS, EXP_BIAS+1, 0x80001000, 0x80000000, TAG_Valid),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x82a0
    },
    {
      MUL,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0
    },
    {
      SUB,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0
    },
    {
      MUL,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0
    },
    {
      MUL,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0
    },
    {
      ADD,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_OVER, 0x00000001, 0x80000000, TAG_Special),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( 0, EXP_BIAS+0xf0f, 0xf0f0f0f0, 0xf0f0f0f0, -1),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x8081
    },
    {
      MUL,
      FULL_PRECISION,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000001, 0x80000001, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000001, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+1, 0x80000001, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0220
    },
    {
      MUL,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000001, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000001, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000003, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x82a0
    },
    {
      DIV,
      FULL_PRECISION,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000003, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000001, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000001, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0020
    },
    {
      DIV,
      FULL_PRECISION,
      REGISTER_( SIGN_POS, EXP_BIAS, 0xfffffffe, 0xbfffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0xfffffffe, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-1, 0xffffffff, 0xbfffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0020
    },
    {
      MUL,
      FULL_PRECISION,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000800, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000800, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00001400, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0020
    },
    {
      DIV,
      FULL_PRECISION,
      REGISTER_( SIGN_POS, EXP_BIAS, 0xfffff000, 0xbfffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0xfffff000, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-1, 0xfffffc00, 0xbfffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0220
    },
    {
      DIV,
      PR_53_BITS,
      REGISTER_( SIGN_POS, EXP_BIAS-0x3ffe, 0x00000000, 0xb0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+0x0010, 0x00000000, 0xf0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+0x1ff1, 0xbbbbb800, 0xbbbbbbbb, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x80b0
    },
    {
      MUL,
      FULL_PRECISION,
      REGISTER_( SIGN_POS, EXP_BIAS+0x3fff, 0xfffffffe, 0xbfffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+0x0010, 0xfffffffe, 0xbfffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_OVER, 0x00000000, 0x80000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0228
    },
    {
      ADD,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_OVER, 0x00000000, 0xc0000100, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_OVER, 0x00000000, 0xc0000100, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0000
    },
    {
      REM1,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS-1, 0xffffffff, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0xfffffffe, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_NEG, EXP_BIAS-1, 0xfffffffd, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0200
    },
    {
      SQRT,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+1, 0x00000000, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0xfffffffe, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0xf9de6484, 0xb504f333, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0020
    },
    {
      SQRT,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS-1, 0xfffefffd, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0xfffffffe, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-1, 0xffff7ffe, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0020
    },
    {
      SCALE,  
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+1, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0000
    },
    {
      SINCOS,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-2, 0x556e10cf, 0xa8ec76d6, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x3820
    },
    {
      SIN,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS-1, 0x00000000, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-1, 0xfc786b2e, 0xae7fe0b5, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0220
    },
    {
      SIN,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+1, 0x00000000, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-3, 0xb6aada7d, 0x9081c36d, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0220
    },
    {
      SIN,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+2, 0x00000000, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_NEG, EXP_BIAS-1, 0xe0f57385, 0xc1bdceee, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0220
    },
    {
      SIN,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+2, 0x00000000, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_NEG, EXP_BIAS-2, 0x851601d7, 0x8f0f8c55, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0220
    },
    {
      SIN,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_NEG, EXP_BIAS-1, 0x00000000, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_NEG, EXP_BIAS-1, 0xfc786b2d, 0xae7fe0b5, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0220
    },
    {
      COS,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0000
    },
    {
      COS,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+2, 0x00000000, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-1, 0xc117abd6, 0xf5cdb84b, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0020
    },
    {
      COS,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0xc0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-4, 0x2fcd3a21, 0x90deaa7e, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0020
    },
    {
      TAN,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x3800
    },
    {
      MULP,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS+0x3fff, 0xf0ffffff, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+2, 0x00000001, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-0x2000+1, 0xf1000001, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x8aa8
    },
    {
      MULP,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS, 0xf0ffffff, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+2, 0x00000001, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+2, 0xf1000001, 0xffffffff, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x8aa0
    },
    {
      MULP,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS+0x3fff, 0x00000f00, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+2, 0x00000f00, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-0x2000+1, 0x00001e00, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x88a8
    },
    {
      MULP,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000f00, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+2, 0x00000f00, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+2, 0x00001e00, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x88a0
    },
    {
      MUL,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS+0x03e4, 0x03ace000, 0xbf21e440, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+0x03e4, 0x03ace000, 0xbf21e440, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+0x07c9, 0x297eff60, 0x8eb39714, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x82a0
    },
    {
      ADD,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_OVER, 0x00000000, 0x80000100, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x00000000, TAG_Zero),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x8081
    },
    {
      ADDI,
      PR_64_BITS | RC_RND,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0xc50b0000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_OVER, 0x00000000, 0x20000100, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0xc50b0000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x8081
    },
    {
      REM1,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER+10, 0xffffffff, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER+10-4, 0x00000000, 0xf0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER+5, 0x00000020, 0xdfffffe0, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0200
    },
    {
      REM1,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0xffffffff, 0x80000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0xf0000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER, 0x00000001, 0x6fffffff, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0202
    },
    {
      REM1,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0xffffffff, 0x80000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x0f000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_NEG, EXP_BIAS+EXP_UNDER, 0x00000001, 0x06ffffff, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0202
    },
    {
      REM,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER+10, 0xffffffff, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER+10, 0x00000000, 0xf0000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER+10, 0xffffffff, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0000
    },
    {
      REM,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0xffffffff, 0x80000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0xf0000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER+1, 0xffffffff, 0x80000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0002
    },
    {
      REM,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0xffffffff, 0x80000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x80000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0xffffffff, 0x00000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0202
    },
    {
      REM,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0xffffffff, 0x80000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x0f000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0xffffffff, 0x08000000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0002
    },
    {
      REM,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER+0x3001, 0x00000000, 0x80001000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-0x40, 0x00000000, 0xfce10000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0400
    },
    {
      REM,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x80001000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-0x40, 0x00000000, 0xfce10000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0402
    },
    {
      REM,
      FULL_PRECISION | CW_Exceptions,
      REGISTER_( SIGN_POS, EXP_BIAS-0x12, 0x00000000, 0x9e065200, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS+EXP_UNDER, 0x00000000, 0x80001000, TAG_Special),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS-0x45, 0x00000000, 0xc7c80000, TAG_Valid),
      REGISTER_( SIGN_POS, EXP_BIAS, 0x00000000, 0x00000000, TAG_Valid),
      REGISTER_( 0, 0, 0, 0, 0), REGISTER_( 0, 0, 0, 0, 0),
      0x0402
    }
  };

struct test native_result;

void ext_mul(void *a, void *b, void *c);
void check(int testnr);
void display_st0(void);
void  complex_test(void);
void test_fprem1(int prec, FPU_REG *s0, FPU_REG *s1, FPU_REG *ok, int stat);
void test_fprem(int prec, FPU_REG *s0, FPU_REG *s1, FPU_REG *ok, int stat);
void check_native(int testnr);
void test_native_sin(int testnr, int prec, FPU_REG *s0, FPU_REG *ok, int stat);

char *symbol[] =
{
  "*",
  "/",
  "+",
  "-",
  "fprem1",
  "fsqrt",
  "fscale",
  "fsincos",
  "cos",
  "tan",
  "mulp",
  "+(i)",
  "fprem",
  "sin"
};

/* finit; fwait; nop */
char init_code[] = { 0xdb, 0xe3, 0x9b, 0x90 };

/* finit; fldcw mem; fwait; nop */
char init_cw_code[] = { 0xdb, 0xe3,
			0xd9, 0x2d, 0x00, 0x00, 0x00, 0x00,
			0x9b, 0x90 };

/* fldl mem; fwait; nop */
char fld_code[] = { 0xdd, 0x05, 0x00, 0x00, 0x00, 0x00, 0x9b, 0x90 };

/* fld mem; fmul mem; fwait; nop */
char fldmul_code[] = { 0xdd, 0x05, 0x00, 0x00, 0x00, 0x00,
		       0xdc, 0x0d, 0x00, 0x00, 0x00, 0x00,
		       0xdd, 0x1d, 0x00, 0x00, 0x00, 0x00,
		       0x9b, 0x90 };

/* fld mem; fdiv mem; fwait; nop */
char flddiv_code[] = { 0xdd, 0x05, 0x00, 0x00, 0x00, 0x00,
		       0xdc, 0x35, 0x00, 0x00, 0x00, 0x00,
		       0xdd, 0x1d, 0x00, 0x00, 0x00, 0x00,
		       0x9b, 0x90 };

/* fld mem; fistp mem; fwait; nop */
char fistp_code[] = { 0xdd, 0x05, 0x00, 0x00, 0x00, 0x00,
		      0xdb, 0x1d, 0x00, 0x00, 0x00, 0x00,
		      0x9b, 0x90 };

double dblx = 0.1, dbly = 30.0, dblz = 0.0;
float fl = 0.2;
long lng;
long double ldblx;


void init_emu(void)
{
  math_arg.___eip = (long)init_code;
  ((longargfunc)math_emulate)(math_arg);
}


void init_emu_cw(int cw)
{
  ((int **)&init_cw_code[4])[0] = &cw;
  math_arg.___eip = (long)init_cw_code;
  ((longargfunc)math_emulate)(math_arg);
}


int
main()
{
  int i;
  char dummy[] = { 0xdd, 0xdd, 0xdd };
  struct task_struct loc_task;

  current = &loc_task;
#if 0
  current->flags = 0;
#endif

  /*  void *ptr = &FPU_info; */

  /* Set a few things to reasonable values. */
  FPU_info = &math_arg;      /* Gets changed in math_emulate */
  FPU_EFLAGS &= ~0x00020000;
  FPU_CS = __USER_CS;
  FPU_DS = __USER_DS;
  FPU_ORIG_EIP = (unsigned)dummy;

  for ( i = 0; i < sizeof(tests)/sizeof(struct test); i++ )
    {
      check_native(i);
      check(i);
    }

  /* Careful here: don't use the emulator macros to access the
     info structure.  In math_emulate() these macros operate on
     a copy which is on the stack as the argument to math_emulate().
     */

  current->used_math = 1;
  current->need_resched = 0;
  current->flags = 0;

#if 0
  current->flags &= ~PF_PTRACED;
#endif

  math_arg.___eip = (long)init_code;
  ((longargfunc)math_emulate)(math_arg);

  ((double **)&fld_code[2])[0] = &dblx;
  math_arg.___eip = (long)fld_code;
  ((longargfunc)math_emulate)(math_arg);
  if ( dblx != 0.1 )
    {
      printk("loaded: %g\n", dblx);
      display_st0();
    }

  ((double **)&fldmul_code[2])[0] = &dblx;
  ((double **)&fldmul_code[8])[0] = &dbly;
  ((double **)&fldmul_code[14])[0] = &dblz;
  math_arg.___eip = (long)fldmul_code;
  ((longargfunc)math_emulate)(math_arg);
  if ( dblz != 3.0 )
    printk("multiply:  %g * %g -> %g\n", dblx, dbly, dblz);

  fldmul_code[0] = 0xd9;
  ((float **)&fldmul_code[2])[0] = &fl;
  ((double **)&fldmul_code[8])[0] = &dbly;
  ((double **)&fldmul_code[14])[0] = &dblz;
  math_arg.___eip = (long)fldmul_code;
  ((longargfunc)math_emulate)(math_arg);
  if ( dblz != fl*dbly )
    printk("multiply:  %g * %g -> %g (diff from 6.0 = %g)\n",
		 fl, dbly, dblz, dblz-6.0);

  init_emu();
  ((double **)&flddiv_code[2])[0] = &dblx;
  ((double **)&flddiv_code[8])[0] = &dbly;
  ((double **)&flddiv_code[14])[0] = &dblz;
  math_arg.___eip = (long)flddiv_code;
  ((longargfunc)math_emulate)(math_arg);
  if ( dblz != 1.0/300.0)
    printk("divide:  %g / %g -> %g\n", dblx, dbly, dblz);

  dblx = 123.0;
  init_emu();
  ((double **)&fistp_code[2])[0] = &dblx;
  ((long **)&fistp_code[8])[0] = &lng;
  math_arg.___eip = (long)fistp_code;
  ((longargfunc)math_emulate)(math_arg);
  if ( lng != 123)
    printk("fistp %g -> %ld\n", dblx, lng);

  ((unsigned short *)&ldblx)[4] = 0x47c8;
  ((unsigned *)&ldblx)[0] = 0x297eff60;
  ((unsigned *)&ldblx)[1] = 0x8eb39714;
  init_emu_cw(0x0322);

  FLDT(*((struct long_double *)&ldblx));
  FSTPL(*((struct long_double *)&dblx));

  if ( dblx != 123.0 )
    printk("fstp %Lg -> %g, status=%04x (wanted b888)\n", ldblx, dblx,
	   (unsigned)status_word());

  complex_test();

  return 0;
}


void display_st0(void)
{
  FPU_REG result;
  char tag, sign;

  reg_copy(&st(0), &result);
  tag = FPU_gettag0();
  sign = getsign_(&st(0));

  printk("st(0) is (%d)%08x%08x(%04x)[%x]\n",
	     sign, result.sigh, result.sigl,
	     (unsigned)(result.exp & 0xffff), tag);
}


double a, b, c;
int native, ii, cw1, cw2;

void  complex_test(void)
{
  a = 1000.1;
  b = 902.03;
  c = 251;
  asm volatile ("fldl %1
	fld %%st(0)
	faddl %2
	fldl %0
	fmul %%st,%%st(1)
	fsubp %%st,%%st(2)
	fdivp %%st,%%st(1)
        fnstcw %3"::"m" (a), "m" (b), "m" (c), "m" (cw1));
  cw2 = (cw1 & 0xff) | 0xc00;
  asm volatile ("fldcw %2\n\tfistpl %0\n\tfldcw %1"
		:"=m" (native): "m" (cw1), "m" (cw2));

  {
    char part1[] = {
      0xdd, 0x05, 0x00, 0x00, 0x00, 0x00,	/* fldl b */
      0xd9, 0xc0,				/* fld %st(0) */
      0xdc, 0x05, 0x00, 0x00, 0x00, 0x00,	/* faddl c */
      0xdd, 0x05, 0x00, 0x00, 0x00, 0x00,	/* fldl a */
      0xdc, 0xc9,				/* fmul %st,%st(1) */
      0xde, 0xe2,				/* fsubp %st,%st(2) */
      0xde, 0xf1,				/* fdivp %st,%st(1) */
      0xd9, 0x3d, 0x00, 0x00, 0x00, 0x00,	/* fnstcw cw1 */
    };
    char part2[] = {
      0xd9, 0x2d, 0x00, 0x00, 0x00, 0x00,	/* fldcw cw2 */
      0xdb, 0x1d, 0x00, 0x00, 0x00, 0x00,	/* fistpl ii */
      0xd9, 0x2d, 0x00, 0x00, 0x00, 0x00,	/* fldcw cw1 */
    };

    init_emu();
    ((double **)&part1[2])[0] = &b;
    ((double **)&part1[10])[0] = &c;
    ((double **)&part1[16])[0] = &a;
    ((int **)&part1[28])[0] = &cw1;
    math_arg.___eip = (long)part1;
    ((longargfunc)math_emulate)(math_arg);

    cw2 = (cw1 & 0xff) | 0xc00;

    ((int **)&part2[2])[0] = &cw2;
    ((int **)&part2[8])[0] = &ii;
    ((int **)&part2[14])[0] = &cw1;
    math_arg.___eip = (long)part2;
    ((longargfunc)math_emulate)(math_arg);

    if ( (native != 11758) || (ii != 11758) )
      {
	printk("Native i = %d,  ", native);
	printk("Emulated i = %d,  ", ii);
	printk("top is %d\n", (int)top);
      }
  }     

  {
    double z = 1.0;
    unsigned x[2] = { 0x8800759c, 0x7e37e43c };
    char ovfl[] = {
      0xdd, 0x05, 0x00, 0x00, 0x00, 0x00,		/*  fldl mem */
      0xdc, 0x0d, 0x00, 0x00, 0x00, 0x00,		/*  fmull mem */
      0xdd, 0x1d, 0x00, 0x00, 0x00, 0x00,		/*  fstpl mem */
      0x90 };						/*  nop */
    ((unsigned **)&ovfl[2])[0] = x;
    ((unsigned **)&ovfl[8])[0] = x;
    ((double **)&ovfl[14])[0] = &z;
    init_emu_cw(0x0322);
    math_arg.___eip = (long)ovfl;
    ((longargfunc)math_emulate)(math_arg);
    if ( z != 1.0 )
      {
	printk("Unmasked overflow -> %g (should be 1.0), "
	       "status = %04x (wanted b8a8)\n",
	       z, (unsigned)status_word());
      }
  }
}


void check_native(int testnr)
{
  int prec, type;
  FPU_REG *a1, *a2, *result, *correct0, *correct1;
  unsigned short wantedswd;

  prec = tests[testnr].control;
  type = tests[testnr].operation;
  a1 = &tests[testnr].arg1;
  a2 = &tests[testnr].arg2;
  result = &tests[testnr].result;
  correct0 = &tests[testnr].correct0;
  correct1 = &tests[testnr].correct1;
  wantedswd = tests[testnr].swd;

  switch ( type )
    {
    case MUL:
    case MULP:
    case DIV:
    case ADD:
    case ADDI:
    case SUB:
      break;
    case REM1:
      test_fprem1(prec, a1, a2, correct0, wantedswd);
      break;
    case REM:
      test_fprem(prec, a1, a2, correct0, wantedswd);
      break;
    case SQRT:
    case SCALE:
    case SINCOS:
    case COS:
      break;
    case SIN:
      test_native_sin(testnr, prec, a1, correct0, wantedswd);
      break;
    case TAN:
      break;
    default:
      reportinterror("Unknown test type  [%d]\n", type);
    }
}


void check(int testnr)
{
  unsigned short statwd, wantedswd;
  int prec, type;
  FPU_REG *a1, *a2, *result, *correct0, *correct1;
  char a1tag, a2tag, corr0tag;
  int nransw = 1, nargs = 2;
  char restag;

  prec = tests[testnr].control;
  type = tests[testnr].operation;
  a1 = &tests[testnr].arg1;
  a1tag = tests[testnr].arg1tag;
  a2 = &tests[testnr].arg2;
  a2tag = tests[testnr].arg2tag;
  result = &tests[testnr].result;
  correct0 = &tests[testnr].correct0;
  corr0tag = tests[testnr].c0tag;
  correct1 = &tests[testnr].correct1;
  wantedswd = tests[testnr].swd;

  finit();
  control_word = prec;

  switch ( type )
    {
    case MUL:
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      reg_copy(a2, &st(1));
      FPU_settagi(1, a2tag);
      copysign(a2, &st(1));
      FPU_rm = 1;
      fmul__();
      reg_copy(&st(0), result);
      copysign(&st(0), result);
      restag = FPU_gettag0();
      break;
    case MULP:
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      reg_copy(a2, &st(1));
      FPU_settagi(1, a2tag);
      copysign(a2, &st(1));
      FPU_rm = 1;
      fmulp_();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case DIV:
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      reg_copy(a2, &st(1));
      FPU_settagi(1, a2tag);
      copysign(a2, &st(1));
      FPU_rm = 1;
      fdiv__();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case ADD:
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      reg_copy(a2, &st(1));
      FPU_settagi(1, a2tag);
      copysign(a2, &st(1));
      FPU_rm = 1;
      fadd__();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case ADDI:
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      reg_copy(a2, &st(1));
      FPU_settagi(1, a2tag);
      copysign(a2, &st(1));
      FPU_rm = 1;
      fadd_i();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case SUB:
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      reg_copy(a2, &st(1));
      FPU_settagi(1, a2tag);
      copysign(a2, &st(1));
      FPU_rm = 1;
      fsub__();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case REM1:
      nransw = 2;
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      reg_copy(a2, &st(1));
      FPU_settagi(1, a2tag);
      copysign(a2, &st(1));
      FPU_rm = 5;
      FPU_triga();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case REM:
      nransw = 2;
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      reg_copy(a2, &st(1));
      FPU_settagi(1, a2tag);
      copysign(a2, &st(1));
      FPU_rm = 0;
      FPU_trigb();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case SQRT:
      nargs = 1;
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      FPU_rm = 2;
      FPU_trigb();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case SCALE:
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      reg_copy(a2, &st(1));
      FPU_settagi(1, a2tag);
      copysign(a2, &st(1));
      FPU_rm = 5;
      FPU_trigb();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case SINCOS:
      nargs = 1;
      nransw = 2;
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      FPU_rm = 3;
      FPU_trigb();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case COS:
      nargs = 1;
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      FPU_rm = 7;
      FPU_trigb();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case SIN:
      nargs = 1;
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      FPU_rm = 6;
      FPU_trigb();
      reg_copy(&st(0), result);
      restag = FPU_gettag0();
      copysign(&st(0), result);
      break;
    case TAN:
      nargs = 1;
      nransw = 2;
      reg_copy(a1, &st(0));
      FPU_settag0(a1tag);
      copysign(a1, &st(0));
      FPU_rm = 2;
      FPU_triga();
      reg_copy(&st(1), result);
      restag = FPU_gettagi(1);
      copysign(&st(1), result);
      break;
    default:
      reportinterror("Unknown test type  [%d]\n", type);
    }
  statwd = status_word();

  if ( (result->sigh != correct0->sigh) || (result->sigl != correct0->sigl)
       || (exponent(result) != exponent(correct0))
       || (restag != corr0tag)
       || (getsign_(result) != getsign_(correct0))
       || (wantedswd != statwd) )
    {
      if ( (wantedswd == statwd) &&
	   (getsign_(result) == getsign_(correct0)) &&
	   (restag == corr0tag) &&
	   (exponent(result) == exponent(correct0)) )
	{
	  long long diff;

	  diff = *(long long *)&result->sigl - *(long long *)&correct0->sigl;
	  if ( (diff <= 2) && (diff >= -2) )
	    printk("Small difference = %d ulp in:\n", (int)diff);
	}
	   
      if ( nargs == 2 )
	printk("(%d)%08x%08x(%04x)[%x] %s (%d)%08x%08x(%04x)[%x]\n",
	       getsign_(a1), a1->sigh, a1->sigl, (unsigned)(a1->exp & 0xffff),
	       a1tag,
	       symbol[type],
	       getsign_(a2), a2->sigh, a2->sigl, (unsigned)(a2->exp & 0xffff),
	       a2tag);
      else
	printk("%s (%d)%08x%08x(%04x)[%x]\n",
	       symbol[type],
	       getsign_(a1), a1->sigh, a1->sigl, (unsigned)(a1->exp & 0xffff),
	       a1tag);
      printk(" -> (%d)%08x%08x(%04x)[%x]   wanted: (%d)%08x%08x(%04x)[%x]\n",
	     getsign_(result), result->sigh, result->sigl,
	     (unsigned)(result->exp & 0xffff), restag,
	     getsign_(correct0), correct0->sigh, correct0->sigl,
	     (unsigned)(correct0->exp & 0xffff), corr0tag
	     );
      if ( wantedswd != statwd )
	printk("status word = %04x, wanted = %04x\n", statwd, wantedswd);
      printk("\n");
      return;
    }

}


void test_native_sin(int testnr, int prec, FPU_REG *s0, FPU_REG *ok, int stat)
{
  long double arg0, result0, result1;
  struct fpu_state all;

  *(long long *)&arg0 = *(long long *)&s0->sigl;
  ((short *)&arg0)[4] = s0->exp;

  asm volatile ("finit; fldcw %0"::"m" (prec));
  asm volatile ("fldt %3; fsin\n"
		"fsave %0; frstor %0\n"
		"fstpt %1; fstpt %2"
		:"=m" (all), "=m" (result0), "=m" (result1)
		:"m" (arg0));
  if ( ((((short *)&result0)[4] & 0x7fff) != (ok->exp & 0x7fff)) ||
       (((long long *)&result0)[0] != ((long long *)&ok->sigl)[0]) ||
       ((((short *)&result0)[4] & 0x8000) != getsign_(ok) << 15) ||
       ((all.status & 0xffff) != ((stat & 0xc7ff) | 0x3000)) )
    {
      printk("Host fsin : %04x %08x%08x\n"
	     "         --> %04x %08x%08x  status=%04x\n",
	     ((short *)&arg0)[4] & 0xffff, ((unsigned *)&arg0)[1],
	     ((unsigned *)&arg0)[0],
	     ((short *)&result0)[4] & 0xffff, ((unsigned *)&result0)[1],
	     ((unsigned *)&result0)[0],
	     all.status & 0xffff);
      printk("     wanted: %04x %08x%08x  status=%04x\n",
	     (int)((ok->exp & 0x7fff) | (getsign_(ok) << 15)),
	     ok->sigh, ok->sigl,
	     (stat & 0xc7ff) | 0x3000);
      printk("\n");
    }
}


void test_fprem(int prec, FPU_REG *s0, FPU_REG *s1, FPU_REG *ok, int stat)
{
  long double arg0, arg1, result0, result1;
  struct fpu_state all;

  *(long long *)&arg0 = *(long long *)&s0->sigl;
  ((short *)&arg0)[4] = s0->exp;
  *(long long *)&arg1 = *(long long *)&s1->sigl;
  ((short *)&arg1)[4] = s1->exp;

  asm volatile ("finit; fldcw %0"::"m" (prec));
  asm volatile ("fldt %3; fldt %4; fprem\n"
		"fsave %0; frstor %0\n"
		"fstpt %1; fstpt %2"
		:"=m" (all), "=m" (result0), "=m" (result1)
		:"m" (arg1), "m" (arg0));
  if ( ((((short *)&result0)[4] & 0x7fff) != (ok->exp & 0x7fff)) ||
       (((long long *)&result0)[0] != ((long long *)&ok->sigl)[0]) ||
       ((((short *)&result0)[4] & 0x8000) != getsign_(ok) << 15) ||
       ((all.status & 0xffff) != ((stat & 0xc7ff) | 0x3000)) )
    {
      printk("Host fprem : %04x %08x%08x  %%  %04x %08x%08x\n"
	     "         --> %04x %08x%08x  status=%04x\n",
	     ((short *)&arg0)[4] & 0xffff, ((unsigned *)&arg0)[1],
	     ((unsigned *)&arg0)[0],
	     ((short *)&arg1)[4] & 0xffff, ((unsigned *)&arg1)[1],
	     ((unsigned *)&arg1)[0],
	     ((short *)&result0)[4] & 0xffff, ((unsigned *)&result0)[1],
	     ((unsigned *)&result0)[0],
	     all.status & 0xffff);
      printk("     wanted: %04x %08x%08x  status=%04x\n",
	     (int)((ok->exp & 0x7fff) | (getsign_(ok) << 15)),
	     ok->sigh, ok->sigl,
	     (stat & 0xc7ff) | 0x3000);
      printk("\n");
    }
}


void test_fprem1(int prec, FPU_REG *s0, FPU_REG *s1, FPU_REG *ok, int stat)
{
  long double arg0, arg1, result0, result1;
  struct fpu_state all;

  *(long long *)&arg0 = *(long long *)&s0->sigl;
  ((short *)&arg0)[4] = s0->exp;
  *(long long *)&arg1 = *(long long *)&s1->sigl;
  ((short *)&arg1)[4] = s1->exp;

  asm volatile ("finit; fldcw %0"::"m" (prec));
  asm volatile ("fldt %3; fldt %4; fprem1\n"
		"fsave %0; frstor %0\n"
		"fstpt %1; fstpt %2"
		:"=m" (all), "=m" (result0), "=m" (result1)
		:"m" (arg1), "m" (arg0));

  if ( ((((short *)&result0)[4] & 0x7fff) != (ok->exp & 0x7fff)) ||
       (((long long *)&result0)[0] != ((long long *)&ok->sigl)[0]) ||
       ((((short *)&result0)[4] & 0x8000) != getsign_(ok) << 15) ||
       ((all.status & 0xffff) != ((stat & 0xc7ff) | 0x3000)) )
    {
      printk("Host fprem1: %04x %08x%08x  %%  %04x %08x%08x\n"
	     "         --> %04x %08x%08x  status=%04x\n",
	     ((short *)&arg0)[4] & 0xffff, ((unsigned *)&arg0)[1],
	     ((unsigned *)&arg0)[0],
	     ((short *)&arg1)[4] & 0xffff, ((unsigned *)&arg1)[1],
	     ((unsigned *)&arg1)[0],
	     ((short *)&result0)[4] & 0xffff, ((unsigned *)&result0)[1],
	     ((unsigned *)&result0)[0],
	     all.status & 0xffff);
      printk("     wanted: %04x %08x%08x  status=%04x\n",
	     (int)((ok->exp & 0x7fff) | (getsign_(ok) << 15)),
	     ok->sigh, ok->sigl,
	     (stat & 0xc7ff) | 0x3000);
      printk("\n");
    }
}
